/* SPDX-FileCopyrightText: 2023 Greenbone AG
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#ifndef MISC_SCANID_H
#define MISC_SCANID_H

int
set_scan_id (const char *);

const char *
get_scan_id (void);
#endif
