/* SPDX-FileCopyrightText: 2023 Greenbone AG
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/*
 * @file openvas.h
 * @brief Headers for OpenVAS entry point.
 */

#ifndef OPENVAS_H
#define OPENVAS_H

int
openvas (int, char **, char **);

#endif /* not OPENVAS_H */
