/* SPDX-FileCopyrightText: 2023 Greenbone AG
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

nav {
    height: 100%;
    width: 20%;
    position: fixed;
    z-index: 1;
    top: 0;
    left: 0;
    overflow-x: hidden;
    padding-top: 20px;
    padding-left: 10px;
    background-color: #dadada;
}

h1 {
    font-size: 25px;
}

h2 {
    font-size: 20px;
}

img {
    max-width: 100%;
    height: auto;
}

nav a {
    text-decoration: none;
    padding: 2px 8px 2px 0px;
    font-size: 18px;
    color: #333333;
    display: block;
}

nav ul {
    display: none;
    padding: 0px 10px;
}

nav>ul,
nav ul.active,
nav ul.active>ul {
    display: block;
}

nav li {
    padding-left: 0px;
}

nav a:hover {
    text-decoration: underline dotted;
}

nav a.selected {
    text-decoration: underline;
}

main {
    height: 100%;
    width: 80%;
    position: fixed;
    overflow-y: scroll;
    z-index: 1;
    top: 0;
    right: 0;
    padding-top: 20px;
    background-color: #ffffff;
}

main article {
    position: relative;
    padding-left: 20px;
    padding-bottom: 20px;
    color: #000000;
}