# SPDX-FileCopyrightText: 2023 Greenbone AG
# Some text descriptions might be excerpted from (a) referenced
# source(s), and are Copyright (C) by the respective right holder(s).
#
# SPDX-License-Identifier: GPL-2.0-or-later

PLUGIN_SET = "202302011009";
PLUGIN_FEED = "Ziggi Di Mik Mik";
FEED_VENDOR = "Greenbone Networks GmbH";
FEED_HOME = "https://www.greenbone.net/en/feed-comparison/";
FEED_NAME = "ZiggiDiMikMik";
