/* SPDX-FileCopyrightText: 2023 Greenbone AG
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#ifndef MISC_STRUTILS_H
#define MISC_STRUTILS_H

#include <glib.h>

int
str_match (const gchar *, const gchar *, int);

#endif
